import abc
import time
import pickle
import logging
import threading
from contextlib import contextmanager
from typing import Generator, Tuple, Any, Optional, Iterator


logger = logging.getLogger(__name__)


class Connection(abc.ABC):
    _HEADER_NBYTES = 4

    def __init__(self) -> None:
        self._cancel_event = threading.Event()
        self._yield_period = 1

    @property
    @abc.abstractmethod
    def input_name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def output_name(self) -> str:
        pass

    def __enter__(self) -> "Connection":
        return self

    def __exit__(self, *_) -> None:
        self.close()

    def cancel(self) -> None:
        self._cancel_event.set()

    def cancelled(self) -> bool:
        return self._cancel_event.is_set()

    @abc.abstractmethod
    def close(self):
        pass

    @abc.abstractmethod
    def wait_client(self) -> None:
        """Wait for the client to be online"""
        pass

    @contextmanager
    def _wait_client_context(self) -> Generator[None, None, None]:
        logger.debug("waiting for remote job to connect to %s ...", self.output_name)
        try:
            yield self._wait_client_iterator()
        except Exception:
            logger.debug(
                "waiting for remote job to connect to %s failed", self.output_name
            )
            raise
        if self.cancelled():
            logger.debug(
                "waiting for remote job to connect to %s cancelled", self.output_name
            )
        else:
            logger.debug("remote job connected to %s", self.output_name)

    def _wait_client_iterator(self) -> Iterator[None]:
        while not self.cancelled():
            yield
            time.sleep(self._yield_period)

    @abc.abstractmethod
    def receive_nbytes(self, nbytes: int) -> bytes:
        pass

    @contextmanager
    def _receive_nbytes_context(self, nbytes: int) -> Generator[None, None, None]:
        buffer = _Buffer(nbytes)

        yield self._receive_nbytes_iterator(buffer, nbytes)

        if len(buffer.data) != nbytes:
            err_msg = f"{len(buffer.data)} bytes received from {self.input_name} instead of {nbytes} bytes"
            if self.cancelled():
                raise ValueError(f"{err_msg} (cancelled)")
            else:
                raise ValueError(err_msg)

    def _receive_nbytes_iterator(
        self, buffer: "_Buffer", nbytes: int
    ) -> Iterator["_Buffer"]:
        while not self.cancelled() and len(buffer.data) < nbytes:
            yield buffer
            time.sleep(self._yield_period)

    @abc.abstractmethod
    def send_bytes(self, data: bytes) -> None:
        pass

    def send_data(self, data: Any) -> None:
        bdata = self._serialize_data(data)
        nbytes = len(bdata)
        logger.debug(
            "send data %s (%d bytes) to client of %s ...",
            type(data),
            nbytes,
            self.output_name,
        )
        bheader = self._serialize_header(bdata)
        try:
            self.send_bytes(bheader + bdata)
        except (BrokenPipeError, ConnectionResetError):
            if data is None:
                logger.debug("client of %s already exited", self.output_name)
                return
            raise
        logger.debug("data send to client of %s", self.output_name)

    def receive_data(self) -> Tuple[Any, Optional[BaseException]]:
        logger.debug("waiting for client data on %s ...", self.input_name)
        bheader = self.receive_nbytes(self._HEADER_NBYTES)
        nbytes = self._deserialize_header(bheader)
        if nbytes:
            logger.debug(
                "receiving %d bytes from client on %s ...", nbytes, self.input_name
            )
        else:
            logger.warning(
                "receiving %d bytes from client on %s (corrupt data?) ...",
                nbytes,
                self.input_name,
            )
        bdata = self.receive_nbytes(nbytes)
        logger.debug("client data received from %s", self.input_name)
        return self._deserialize_data(bdata)

    def _serialize_header(self, bdata: bytes) -> bytes:
        return len(bdata).to_bytes(self._HEADER_NBYTES, "big")

    def _deserialize_header(self, bheader: bytes) -> int:
        return int.from_bytes(bheader, "big")

    def _serialize_data(self, data: Any) -> bytes:
        return pickle.dumps(data)

    def _deserialize_data(self, data: bytes) -> Any:
        return pickle.loads(data)


class _Buffer:
    def __init__(self, nbytes: int) -> None:
        self.data = b""
        self._nbytes = nbytes
        self._max_chunk_size = min(nbytes, 512)

    @property
    def chunk_size(self) -> int:
        return min(self._max_chunk_size, self._nbytes - len(self.data))



import os
import sys
import json
import logging
import pathlib
import subprocess
from typing import Optional, Dict, Tuple, Union, List
from unittest.mock import patch
from tempfile import NamedTemporaryFile
from contextlib import contextmanager
from concurrent.futures import ThreadPoolExecutor

import pytest
import requests

from ..client import defaults
from ..client.rest.api import slurm_access

logger = logging.getLogger(__name__)

_MAX_JOBS = 50


@contextmanager
def mock_slurm_clients(tmp_path: pathlib.Path, slurm_parameters: Dict[str, str]):
    last_job_id = 0
    jobs = dict()
    base_url = slurm_parameters["url"]
    api_version_str = slurm_parameters["api_version"] or defaults.DEFAULT_API_VERSION
    api_version_tuple = slurm_access.parse_version(api_version_str)

    class MockResponse(requests.Response):
        def __init__(self, json_data, status_code: int):
            super().__init__()
            self.encoding = "utf-8"
            self._content = json.dumps(json_data).encode(self.encoding)
            self.status_code = status_code
            self.headers["Content-Type"] = "application/json"

    def _mock_request(
        method: str, url: str, json: Optional[dict] = None, **_
    ) -> MockResponse:
        nonlocal last_job_id
        path = url[len(base_url) :]

        if (method, path) == ("GET", "/openapi"):
            return MockResponse(
                {
                    "paths": {
                        f"/slurm/{api_version_str}/job/submit": {},
                    }
                },
                200,
            )

        if method == "GET" and path.startswith(f"/slurm/{api_version_str}/job/"):
            job_id = int(path.split("/")[-1])
            job_exec_info = jobs.get(job_id)
            if job_exec_info is None:
                return MockResponse(
                    {
                        "last_backfill": {"set": False, "infinite": False, "number": 0},
                        "last_update": {"set": False, "infinite": False, "number": 0},
                        "warnings": [],
                        "errors": [],
                    },
                    404,
                )
            logger.debug("SLURM mock backend: get job %s", job_id)
            return MockResponse(
                {
                    "jobs": [job_exec_info],
                    "last_backfill": {"set": False, "infinite": False, "number": 0},
                    "last_update": {"set": False, "infinite": False, "number": 0},
                    "warnings": [],
                    "errors": [],
                },
                200,
            )

        if method == "GET" and path == f"/slurm/{api_version_str}/jobs":
            logger.debug("SLURM mock backend: get jobs (# %d)", len(jobs))
            return MockResponse(
                {
                    "jobs": list(jobs.values()),
                    "last_backfill": {"set": False, "infinite": False, "number": 0},
                    "last_update": {"set": False, "infinite": False, "number": 0},
                    "warnings": [],
                    "errors": [],
                },
                200,
            )

        if method == "GET":
            raise NotImplementedError(path)

        if method == "DELETE" and path.startswith(f"/slurm/{api_version_str}/job/"):
            job_id = int(path.split("/")[-1])
            job_exec_info = jobs.get(job_id)
            if job_exec_info is None:
                return MockResponse(
                    {
                        "warnings": [],
                        "errors": [],
                    },
                    404,
                )
            else:
                job_exec_info["job_state"] = _job_state("CANCELLED", api_version_tuple)
                logger.debug("SLURM mock backend: cancel job %s", last_job_id)
                return MockResponse(
                    {
                        "warnings": [],
                        "errors": [],
                    },
                    200,
                )

        if method == "DELETE":
            raise NotImplementedError(path)

        if path == f"/slurm/{api_version_str}/job/submit":
            last_job_id += 1
            job_exec_info = {
                "job_id": last_job_id,
                "job_state": _job_state("PENDING", api_version_tuple),
                **json["job"],
                "user_name": slurm_parameters["user_name"],
                "warnings": [],
                "errors": [],
            }
            jobs[last_job_id] = job_exec_info
            logger.debug("SLURM mock backend: recieved job %s", last_job_id)

            if api_version_tuple < (0, 0, 39):
                script = json["script"]
            else:
                script = json["job"]["script"]

            lines = script.split("\n")
            shebang = lines[0]
            if "bash" in shebang:
                if sys.platform == "win32":
                    pytest.skip("bash script does not run on windows")
            elif "python" in shebang:
                pass
            else:
                assert False, f"Unknown script starting with '{shebang}'"

            pool.submit(
                _job_main, script, json, job_exec_info, str(tmp_path), api_version_tuple
            )
            return MockResponse(job_exec_info, 200)

        if method == "POST":
            raise NotImplementedError(path)

    with ThreadPoolExecutor(max_workers=_MAX_JOBS) as pool:
        with patch("requests.request", side_effect=_mock_request):
            yield


def _job_main(
    script: str,
    json: dict,
    job_exec_info: dict,
    tmp_path: pathlib.Path,
    api_version: Tuple[int, int, int],
) -> None:
    cmd = []
    lines = script.split("\n")
    shebang = lines[0]
    if "bash" in shebang:
        suffix = ".sh"
    elif "python" in shebang:
        if sys.platform == "win32":
            lines.pop(0)
            cmd = [sys.executable]
        suffix = ".py"
    else:
        assert False, f"Unknown script starting with '{shebang}'"

    with NamedTemporaryFile("w", delete=False, dir=tmp_path, suffix=suffix) as script:
        script.write("\n".join(lines))
        filename = script.name
    os.chmod(filename, 0o755)
    cmd.append(filename)

    env = dict(os.environ)
    for keyvalue in json["job"].get("environment", list()):
        key, _, value = keyvalue.partition("=")
        env[key] = value
    env["SLURM_JOB_ID"] = str(job_exec_info["job_id"])

    standard_output = json["job"].get("standard_output")
    standard_error = json["job"].get("standard_error")

    if standard_output is None:
        stdout = None
    elif standard_output == "/dev/null":
        stdout = None
        standard_output = None
    else:
        stdout = subprocess.PIPE

    if standard_error is None:
        stderr = None
    elif standard_error == "/dev/null":
        stderr = None
        standard_error = None
    else:
        stderr = stdout or subprocess.PIPE

    with subprocess.Popen(
        cmd, stdout=stdout, stderr=stderr, env=env, cwd=os.getcwd()
    ) as proc:
        logger.debug("SLURM mock backend: job %s started", job_exec_info["job_id"])
        job_exec_info["job_state"] = _job_state("RUNNING", api_version)
        outs, errs = proc.communicate(timeout=15)

        if standard_output is not None:
            outfile = standard_output.replace("%j", str(job_exec_info["job_id"]))
            with open(outfile, "wb") as f:
                f.write(outs)

        if standard_error is not None:
            errfile = standard_error.replace("%j", str(job_exec_info["job_id"]))
            with open(errfile, "wb") as f:
                f.write(errs)

        if job_exec_info["job_state"] != _job_state("CANCELLED", api_version):
            if proc.returncode:
                job_exec_info["job_state"] = _job_state("FAILED", api_version)
                logger.debug(
                    "SLURM mock backend: job %s failed", job_exec_info["job_id"]
                )
            else:
                job_exec_info["job_state"] = _job_state("COMPLETED", api_version)
                logger.debug(
                    "SLURM mock backend: job %s completed", job_exec_info["job_id"]
                )


def _job_state(state: str, api_version: Tuple[int, int, int]) -> Union[str, List[str]]:
    if api_version < (0, 0, 40):
        return state
    return [state]



import os
import time
import logging

from ._base import Client
from ._base import main_task_loop
from ._base import init_job


def main():
    init_job()
    with FileClient() as client:
        main_task_loop(client)


class FileClient(Client):
    def __init__(self) -> None:
        input_filename = os.environ["_PYSLURMUTILS_INFILE"]
        output_filename = os.environ["_PYSLURMUTILS_OUTFILE"]

        logging.debug("Connecting to '%s' ...", input_filename)
        input_dirname = os.path.dirname(input_filename)
        while True:
            try:
                _ = os.listdir(input_dirname)  # force NFS cache
                self._input_file = open(input_filename, "rb+")
                break
            except FileNotFoundError:
                time.sleep(0.5)
        logging.debug("Connected to '%s'", input_filename)

        self._output_file = open(output_filename, "wb+")
        self._output_filename = output_filename

        super().__init__(input_filename, output_filename)

    def close(self) -> None:
        if self._input_file is not None:
            self._input_file.close()
            self._input_file = None
        if self._output_file is not None:
            self._output_file.close()
            self._output_file = None

    def _send_bytes(self, bdata):
        self._output_file.write(bdata)
        self._output_file.flush()

    def _receive_nbytes(self, nbytes):
        data = b""
        block = min(nbytes, 512)
        while len(data) < nbytes:
            data += self._input_file.read(block)
            time.sleep(0.1)
        return data



import datetime
from typing import Tuple, Any, Dict, Union, List

_VersionType = Tuple[int, int, int]


def timedelta_in_minutes(version: _VersionType, delta_time: Any) -> Any:
    if isinstance(delta_time, str):
        try:
            hours, minutes, seconds = map(int, delta_time.split(":"))
            delta = datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
        except Exception:
            # Wrong string format. Try submitting the Slurm job anyway.
            pass
        else:
            delta_time = delta.total_seconds() / 60.0
    if isinstance(delta_time, int):
        return integral_number(version, delta_time)
    if isinstance(delta_time, float):
        return integral_number(version, delta_time + 0.5)
    return delta_time


def environment(
    version: _VersionType, environment: Dict[str, Any]
) -> Union[Dict[str, str], List[str]]:
    if version < (0, 0, 39):
        return {k: str(v) for k, v in environment.items()}
    else:
        return [f"{k}={v}" for k, v in environment.items()]


def integral_number(version: _VersionType, number: Any) -> Any:
    if not isinstance(number, int):
        try:
            number = int(number)
        except Exception:
            return number
    return {"number": number, "set": True, "infinite": False}



import logging
import weakref
from pprint import pformat
from concurrent import futures
from contextlib import ExitStack
from contextlib import contextmanager
from typing import Optional, Any

from ..client import errors
from ..client import defaults
from ..client import SlurmPyConnRestClient
from ..client.rest.api import slurm_response
from ..client.job_io.local import RemoteExecutor
from ..client.job_io.local import ExecuteContextReturnType
from ..client.job_io.local import RemoteWorkerProxy
from ..client.job_io.local import FileConnection
from ..client.job_io.local import TcpConnection

logger = logging.getLogger(__name__)


class SlurmRestFuture(futures.Future):
    def __init__(self) -> None:
        self._job_id = None
        self._slurm_client = None
        self._delayed_cancel_job = False
        super().__init__()

    def job_submitted(self, job_id: int, slurm_client: SlurmPyConnRestClient) -> None:
        """The SLURM job was submitted. Beware that the Slurm job may be running other tasks as well."""
        self._job_id = job_id
        self._slurm_client = weakref.proxy(slurm_client)
        if self._delayed_cancel_job:
            slurm_client.cancel_job(job_id)

    @property
    def job_id(self) -> Optional[int]:
        return self._job_id

    @property
    def slurm_client(self) -> Optional[SlurmPyConnRestClient]:
        return self._slurm_client

    def cancel_job(self) -> None:
        """Cancel the Slurm job, not the future. Beware that the Slurm job may be running other tasks as well."""
        slurm_client = self.slurm_client

        # The SLURM job was asked to be cancelled but didn't start yet: use `_delayed_cancel_job` to cancel it after it started.
        if slurm_client is None:
            self._delayed_cancel_job = True
        else:
            slurm_client.cancel_job(self.job_id)


class SlurmRestExecutor(RemoteExecutor):
    _FUTURE_CLASS = SlurmRestFuture

    def __init__(
        self,
        url: str = "",
        user_name: str = "",
        token: str = "",
        api_version: str = "",
        parameters: Optional[dict] = None,
        log_directory: Optional[str] = None,
        std_split: Optional[bool] = False,
        request_options: Optional[dict] = None,
        pre_script: Optional[str] = None,
        post_script: Optional[str] = None,
        python_cmd: Optional[str] = None,
        initializer: Optional[callable] = None,
        initargs: Optional[tuple] = None,
        initkwargs: Optional[tuple] = None,
        data_directory: Optional[str] = None,
        max_workers: Optional[int] = None,
        max_tasks_per_worker: Optional[int] = 1,
        lazy_scheduling: bool = True,
        conservative_scheduling: bool = False,
        cleanup_job_artifacts: bool = True,
    ):
        """
        :param url: SLURM REST API
        :param user_name: User name on SLURM
        :param token: SLURM access token for the user
        :param api_version: SLURM API version.
        :param parameters: SLURM job parameters
        :param log_directory: SLURM log directory
        :param std_split: Split standard output and standard error
        :param request_options: GET, POST and DELETE options
        :param pre_script: Shell script to execute at the start of a job
        :param post_script: Shell script to execute at the end of a job
        :param python_cmd: Python command
        :param initializer: execute when starting a job
        :param initargs: parameters for `initializer`
        :param initkwargs: parameters for `initializer`
        :param data_directory: communicate with the Slum job through files when specified
        :param max_workers: maximum number of Slum jobs that can run at any given time. `None` means unlimited.
        :param max_tasks_per_worker: maximum number of tasks each Slum job can receive before exiting. `None` means unlimited.
        :param lazy_scheduling: schedule SLURM jobs only when needed. Can only be disabled when `max_workers` is specified.
        :param conservative_scheduling: schedule the least amount of workers at the expense of tasks staying longer in the queue.
        :param cleanup_job_artifacts: cleanup job artifacts like logs.
        """

        self._proxy_kwargs = {
            "max_tasks": max_tasks_per_worker,
            "initializer": initializer,
            "initargs": initargs,
            "initkwargs": initkwargs,
        }

        if data_directory:
            self._file_connection_kwargs = {
                "directory": data_directory,
                "basename": defaults.JOB_NAME,
            }
        else:
            self._file_connection_kwargs = None

        self._cleanup_job_artifacts = cleanup_job_artifacts

        self._slurm_client = SlurmPyConnRestClient(
            url=url,
            user_name=user_name,
            token=token,
            api_version=api_version,
            log_directory=log_directory,
            parameters=parameters,
            std_split=std_split,
            request_options=request_options,
            pre_script=pre_script,
            post_script=post_script,
            python_cmd=python_cmd,
        )

        super().__init__(
            max_workers=max_workers,
            max_tasks_per_worker=max_tasks_per_worker,
            lazy_scheduling=lazy_scheduling,
            conservative_scheduling=conservative_scheduling,
        )

    @contextmanager
    def execute_context(self) -> ExecuteContextReturnType:
        if self._file_connection_kwargs:
            conn_ctx = FileConnection(**self._file_connection_kwargs)
        else:
            conn_ctx = TcpConnection()

        with ExitStack() as stack:
            connection = stack.enter_context(conn_ctx)
            proxy_ctx = RemoteWorkerProxy(connection, **self._proxy_kwargs)
            worker_proxy = stack.enter_context(proxy_ctx)

            job_id = None
            first_submit_kw = None

            def initialize(submit_kw):
                """
                Initialize SLURM worker: submit the SLURM job and initialize
                the communication with the worker.
                """
                nonlocal job_id, first_submit_kw
                first_submit_kw = submit_kw
                if submit_kw is None:
                    submit_kw = dict()
                job_id = self._slurm_client.submit_script(worker_proxy, **submit_kw)
                log_ctx = self._slurm_client.redirect_stdout_stderr(job_id)
                _ = stack.enter_context(log_ctx)

                if self._cleanup_job_artifacts:
                    cleanup_ctx = self._slurm_client.clean_job_artifacts_context(job_id)
                    _ = stack.enter_context(cleanup_ctx)

                worker_proxy.initialize()

            def execute(
                task: callable, args: tuple, kwargs: dict, future: SlurmRestFuture
            ) -> Any:
                """
                Send a task to the SLURM worker. Start the worker when not already running.
                """
                nonlocal job_id, first_submit_kw

                submit_kw = kwargs.pop(defaults.SLURM_ARGUMENTS_NAME, None)
                if job_id is None:
                    initialize(submit_kw)
                elif submit_kw != first_submit_kw:
                    logger.warning(
                        "SLURM submit arguments\n %s\n are ignored in favor of the client arguments\n %s",
                        pformat(submit_kw),
                        pformat(first_submit_kw),
                    )

                future.job_submitted(job_id, self._slurm_client)

                # TODO: race condition = job goes down between this check and
                #       to reception of the task. After the job received the task
                #       it should be ok.
                exit_msg = worker_exit_msg()
                if exit_msg:
                    raise errors.RemoteExit(exit_msg) from None

                try:
                    result, exc_info = worker_proxy.execute_without_reraise(
                        task, args=args, kwargs=kwargs
                    )
                except Exception:
                    exit_msg = worker_exit_msg()
                    if exit_msg:
                        raise errors.RemoteExit(exit_msg) from None
                    raise

                if exc_info is not None:
                    errors.reraise_remote_exception_from_tb(exc_info)
                return result

            def worker_exit_msg() -> Optional[str]:
                """
                Worker exit message in case the worker (SLURM job) was running
                but is no longer running.
                """
                if job_id is None:
                    return
                try:
                    status = self._slurm_client.get_status(job_id)
                except Exception as e:
                    logger.warning("failed getting the job state: %s", e, exc_info=True)
                    status = None
                if status not in slurm_response.FINISHING_STATES:
                    return
                return f"SLURM job {job_id} {status}"

            if not self._lazy_scheduling:
                try:
                    initialize(None)
                except Exception as e:
                    logger.warning(
                        "SLURL worker initialization failed: %s", e, exc_info=True
                    )

            yield (execute, worker_exit_msg)



import os
import socket
import pathlib
from unittest.mock import patch
from typing import Optional, Generator

import pytest

from .mock_slurm import mock_slurm_clients


def pytest_addoption(parser):
    parser.addoption(
        "--slurm-root-directory",
        action="store",
        default=None,
        help="Specify the SLURM root directory for logs and data.",
    )
    parser.addoption(
        "--slurm-api-version",
        action="store",
        default=None,
        help="Specify the SLURM API version.",
    )


@pytest.fixture(scope="session")
def slurm_root_directory(tmp_path_factory, request):
    path = request.config.getoption("--slurm-root-directory")
    if path:
        return pathlib.Path(path)
    return tmp_path_factory.mktemp("slurm_root")


@pytest.fixture(scope="session", params=["mock", "production"])
def slurm_parameters(request) -> dict:
    api_version = request.config.getoption("--slurm-api-version")

    if request.param == "mock":
        return {
            "mock": True,
            "url": "mockurl",
            "token": "mocktoken",
            "user_name": "mockuser",
            "api_version": api_version,
        }

    if request.param == "production":
        url = os.environ.get("SLURM_URL")
        if not url:
            pytest.skip("SLURM_URL environment variable required")

        token = os.environ.get("SLURM_TOKEN")
        if not token:
            pytest.skip("SLURM_TOKEN environment variable required")

        user_name = os.environ.get("SLURM_USER")
        if not user_name:
            pytest.skip("SLURM_USER environment variable required")

        slurm_root_directory = request.config.getoption("--slurm-root-directory")
        if not slurm_root_directory:
            pytest.skip("--slurm-root-directory pytest argument is required")

        api_version = os.environ.get("SLURM_API_VERSION", api_version)

        return {
            "mock": False,
            "url": url,
            "token": token,
            "user_name": user_name,
            "api_version": api_version,
        }

    raise ValueError(request.param)


@pytest.fixture(scope="session")
def slurm_log_directory(
    slurm_root_directory, tmp_path_factory, slurm_parameters
) -> Optional[str]:
    if slurm_parameters["mock"]:
        return tmp_path_factory.mktemp("slurm_logs")
    if slurm_root_directory:
        return slurm_root_directory / slurm_parameters["user_name"] / "slurm_logs"


@pytest.fixture(scope="session")
def slurm_data_directory(
    slurm_root_directory, tmp_path_factory, slurm_parameters
) -> Optional[str]:
    if slurm_parameters["mock"]:
        return tmp_path_factory.mktemp("slurm_data")
    if slurm_root_directory:
        return slurm_root_directory / slurm_parameters["user_name"] / "slurm_data"


@pytest.fixture(scope="session")
def mock_job_name():
    job_name = f"pyslurmutils.unittest.{socket.gethostname()}"
    with patch("pyslurmutils.client.defaults.JOB_NAME", job_name):
        yield job_name


@pytest.fixture(scope="session")
def slurm_client_kwargs(
    slurm_log_directory, tmp_path_factory, slurm_parameters, mock_job_name
) -> Generator[dict, None, None]:
    # This must remain a session scope fixture for third-party libraries
    params = dict(slurm_parameters)
    _ = params.pop("mock")
    params["log_directory"] = slurm_log_directory
    if slurm_parameters["mock"]:
        tmp_path = tmp_path_factory.mktemp("slurm_mock")
        with mock_slurm_clients(tmp_path, slurm_parameters):
            yield params
    else:
        yield params
